"""
Environment config
"""
import importlib.util
import json
import typing as t
from locale import getpreferredencoding
from pathlib import Path

import pkg_resources
from pydantic import BaseSettings


class Settings(BaseSettings):
    """
    Env Configuration
    """

    API_V1_STR: str = "/api/v1"

    DEBUG: bool = False
    WEB_URL: str = "https://app.taktile.com"
    TAKTILE_API_URL: str = "https://taktile-api.taktile.com"
    DEPLOYMENT_API_URL: str = "https://deployment-api.taktile.com"

    TAKTILE_API_KEY: t.Optional[str] = None
    TAKTILE_CONFIG_FILE: Path = Path("~/.config/tktl/config.json").expanduser()

    # Used to set the size of each arrow batch in MB. This setting
    # gets overridden if batch size is set via `nrows` property.
    ARROW_BATCH_MB: float = 5

    # Should this client support communication to prereleases?
    ALLOW_PRERELEASE: bool = False

    class Config:  # pylint: disable=too-few-public-methods
        """Env config"""

        case_sensitive = True


settings = Settings()

arrow_available = (
    importlib.util.find_spec("pandas") is not None
    and importlib.util.find_spec("numpy") is not None
    and importlib.util.find_spec("pyarrow") is not None
    and importlib.util.find_spec("certifi") is not None
)

version = pkg_resources.get_distribution(
    __name__.split(".", maxsplit=1)[0]
).version


def collect_api_key() -> t.Optional[str]:
    """collect_api_key
    If an api key is available in the TAKTILE_API_KEY environment variable,
    that one will be used. Otherwise, the taktile config file is read and
    the api key is obtained from there.

    Returns
    -------
    t.Optional[str]
        the API key, if found, else None
    """
    if settings.TAKTILE_API_KEY:
        return settings.TAKTILE_API_KEY

    if settings.TAKTILE_CONFIG_FILE.exists():
        encoding = getpreferredencoding(False)
        with settings.TAKTILE_CONFIG_FILE.open("r", encoding=encoding) as cfg:
            config = json.load(cfg)
        if "api-key" in config:
            return str(config["api-key"])
    return None



"""
Endpoint implementation for the rest client.
"""
import typing as t

from pyarrow.flight import ActionType, FlightClient, Ticket  # type: ignore
from pydantic import BaseModel

from taktile_client.arrow.serialize import deserialize
from taktile_client.arrow.utils import batch_arrow


class EndpointActionGroup(BaseModel):
    """
    Actiongroup of an Arrow endpoint.

    Collects the endpoint and sample data actions into a single type
    """

    endpoint: ActionType
    X: ActionType
    y: ActionType


class ClientArrowEndpoint:
    """
    The actual endpoint object.

    This is an Arrow endpoint, exposing function calling through
    `__call__`. This is the object that is `client.endpoints.repayment`.

    Parameters
    ----------
    client : FlightCLient
        the client to use for making the calls
    action_group : EndpointActionGroup
        the endpoint action_group for this endpoint
    """

    def __init__(
        self, *, client: FlightClient, action_group: EndpointActionGroup
    ):
        self._client = client
        self._action_group = action_group
        self.__name__ = action_group.endpoint.type

    def __call__(self, payload: t.Any, nrows: t.Optional[int] = None) -> t.Any:
        """
        Call the endpoint

        Parameters
        ----------
        payload : t.Any
            payload to send

        nrows : t.Optional[int]
            number of rows to send per batch
        Returns
        -------
        t.Any
            The response of the endpoint
        """
        return batch_arrow(
            client=self._client,
            action=self._action_group.endpoint,
            payload=payload,
            nrows=nrows,
        )

    def X(self) -> t.Any:  # pylint: disable=invalid-name
        """
        Sample input data of endpoint

        Returns
        -------
        t.Any
            sample data
        """
        reader = self._client.do_get(Ticket(ticket=self._action_group.X.type))
        return deserialize(reader.read_all())

    def y(self) -> t.Any:  # pylint: disable=invalid-name
        """
        Sample output data of endpoint

        Returns
        -------
        t.Any
            sample data
        """
        reader = self._client.do_get(Ticket(ticket=self._action_group.y.type))
        return deserialize(reader.read_all())


class ArrowEndpoints:  # pylint: disable=too-few-public-methods
    """
    A collection object for Arrow endpoints.

    This is the `endpoints` object on Taktile Arrow Clients.

    Parameters
    ----------
    client : FlightClient
        the client to use for calls
    actions : t.List[ActionType]
        ungrouped actions of model
    """

    def __init__(self, *, client: FlightClient, actions: t.List[ActionType]):

        action_types = [a.type for a in actions]
        action_groups = []

        for action in actions:

            if (
                action.type + "__X" in action_types
                and action.type + "__y" in action_types
            ):
                action_groups.append(
                    EndpointActionGroup(
                        endpoint=action,
                        X=[
                            a for a in actions if a.type == action.type + "__X"
                        ][0],
                        y=[
                            a for a in actions if a.type == action.type + "__y"
                        ][0],
                    )
                )

        for group in action_groups:
            setattr(
                self,
                group.endpoint.type,
                ClientArrowEndpoint(
                    client=client,
                    action_group=group,
                ),
            )



"""
Base classes for Rest and Arrow client
"""
import typing as t

from taktile_client.config import collect_api_key, settings
from taktile_client.deployment_api import DeploymentApiClient

EndpointClass = t.TypeVar("EndpointClass")


class ApiKeyException(Exception):
    """Exception related to problems with the API key"""


class Client(t.Generic[EndpointClass]):
    """
    Abstract base client for talking to models deployed with Taktile.

    Parameters
    ----------
    api_key : str
        api_key to use
    repository_name : str
        repository_name of the form "taktile-org/repo-name"
    branch_name : str
        branch_name either "main" or "refs/heads/main"
    """

    def __init__(
        self,
        repository_name: str,
        branch_name: str,
        api_key: t.Optional[str] = None,
        headers: t.Optional[t.Dict[str, str]] = None,
    ):
        if api_key is None:
            api_key = collect_api_key()

        if not api_key and not settings.DEBUG:
            raise ApiKeyException(
                "No API key found in environment (TAKTILE_API_KEY) or "
                f"config file ({settings.TAKTILE_CONFIG_FILE.absolute()}). "
                "Please specify an API key when instantiating the client."
            )

        self._api_key = api_key
        principal, repository = repository_name.split("/")
        client = DeploymentApiClient(api_key=api_key, headers=headers)
        host = client.get_host(
            principal=principal, repository=repository, branch=branch_name
        )

        url = self._url_from_host(host)

        self._endpoints = self._initalize_endpoints(url, api_key)

    def _url_from_host(self, host: str) -> str:
        raise NotImplementedError

    def _initalize_endpoints(
        self, url: str, api_key: t.Optional[str]
    ) -> EndpointClass:
        raise NotImplementedError

    @classmethod
    def from_url(
        cls: t.Type["Client[t.Any]"], url: str, api_key: t.Optional[str]
    ) -> "Client[t.Any]":
        """
        Instantiate client from url

        Parameters
        ----------
        api_key : str
            api_key to use
        url : str
            url where the model is deployed

        Returns
        -------
        Client

        """
        obj = t.cast("Client[t.Any]", cls.__new__(cls))
        obj._endpoints = obj._initalize_endpoints(url, api_key)
        return obj

    @property
    def endpoints(self) -> EndpointClass:
        """
        The arrow endpoints

        Returns
        -------
        ArrowEndpoints

        """
        return self._endpoints



"""
Pydantic rest serialization
"""


import pydantic


def pydantic_model_serialize(*, value: pydantic.BaseModel) -> str:
    """
    Serialize pandas series

    Parameters
    ----------
    value : pydantic.BaseModel
        value to serialize

    Returns
    -------
    str
        serialized pydantic model
    """
    return value.json()



"""
Serialize / Deserialize implementations
"""

import logging
from collections.abc import Sequence
from functools import singledispatch
from typing import Any, Type

import numpy
import pandas  # type: ignore
import pyarrow  # type: ignore

import taktile_client.arrow.serialize.serializer as arrow
from taktile_client.exceptions import SerializationError

from .serializer import ObjectSerializer

logger = logging.getLogger(__name__)


def deserialize_arrow(model_input: pyarrow.Table) -> Any:
    """deserialize for arrow"""
    if b"CLS" not in model_input.schema.metadata:
        as_pandas = model_input.to_pandas()
        if len(as_pandas.columns) == 1:
            return as_pandas.squeeze()
        return as_pandas
    cls_name = model_input.schema.metadata[b"CLS"]
    func = getattr(getattr(arrow, cls_name.decode("utf-8")), "deserialize")
    return func(model_input)


@singledispatch
def serialize_arrow(model_input: Any) -> pyarrow.Table:
    """serialize for arrow"""
    return arrow.BinarySerializer.serialize(model_input)


@serialize_arrow.register(Sequence)
def _sequence(model_input: Any) -> pyarrow.Table:
    return _do_serialize_with_fallback(
        arrow.SequenceSerializer, model_input=model_input
    )


@serialize_arrow.register(dict)
def _dict(model_input: Any) -> pyarrow.Table:
    return _do_serialize_with_fallback(
        arrow.SequenceSerializer, model_input=model_input
    )


@serialize_arrow.register(numpy.ndarray)
def _ndarry(model_input: Any) -> pyarrow.Table:
    return _do_serialize_with_fallback(
        arrow.ArraySerializer, model_input=model_input
    )


@serialize_arrow.register(pandas.DataFrame)
def _df(model_input: Any) -> pyarrow.Table:
    return _do_serialize_with_fallback(
        arrow.DataFrameSerializer, model_input=model_input
    )


@serialize_arrow.register(pandas.Series)
def _series(model_input: Any) -> pyarrow.Table:
    return _do_serialize_with_fallback(
        arrow.SeriesSerializer, model_input=model_input
    )


def _do_serialize_with_fallback(
    serializer_cls: Type[ObjectSerializer], model_input: Any
) -> pyarrow.Table:
    try:
        return serializer_cls.serialize(value=model_input)
    except SerializationError:
        logger.warning(
            "Invalid serialization value, "
            "will use binary representation for object"
        )
    return arrow.BinarySerializer.serialize(model_input)



"""
REST serialization
"""
import typing as t
from functools import singledispatch

import pydantic

from taktile_client.config import arrow_available
from taktile_client.exceptions import SerializationError

from .pydantic_model import pydantic_model_serialize
from .utils import JSONStructure

if arrow_available:
    import numpy as np
    import numpy.typing as npt
    import pandas as pd  # type: ignore

    from .dataframe import (
        dataframe_deserialize,
        dataframe_serialize,
        dataframe_to_example,
    )
    from .numpy import numpy_deserialize, numpy_serialize, numpy_to_example
    from .series import series_deserialize, series_serialize, series_to_example


@singledispatch
def serialize(value) -> str:  # type: ignore
    """
    Serialize object for REST

    Parameters
    ----------
    value
        value to serialize

    Returns
    -------
    str
        serialized object
    """
    raise SerializationError(
        f"""Can't serialize value of type {type(value)}.
        You may need to install extra requirements."""
    )


@singledispatch
def deserialize(sample, *, value: JSONStructure):  # type: ignore
    """
    Deserialize object

    Parameters
    ----------
    sample : InputType
        sample object
    value : JSONStructure
        value to deserialize

    Returns
    -------
    InputType
        deserialized object
    """
    raise SerializationError(f"Can't deserialize value of type {type(sample)}")


@singledispatch
def to_example(value) -> str:  # type: ignore
    """
    Produce serialized example of object

    Parameters
    ----------
    value: InputType
        value to produce an example of

    Returns
    -------
    str
        serialized example
    """
    raise SerializationError(
        f"Can't create example value of type {type(value)}"
    )


@serialize.register
def _serialize_pydantic_model(value: pydantic.BaseModel) -> str:
    return pydantic_model_serialize(value=value)


if arrow_available:

    @serialize.register
    def _serialize_numpy(value: np.ndarray) -> str:  # type: ignore
        return numpy_serialize(value=value)

    @serialize.register
    def _serialize_series(value: pd.Series) -> str:
        return series_serialize(value=value)

    @serialize.register
    def _serialize_dataframe(value: pd.DataFrame) -> str:
        return dataframe_serialize(value=value)

    @deserialize.register
    def _deserialize_numpy(
        sample: np.ndarray, *, value: JSONStructure  # type: ignore
    ) -> npt.NDArray[t.Any]:
        return numpy_deserialize(value=value, sample=sample)

    @deserialize.register
    def _deserialize_series(
        sample: pd.Series, *, value: JSONStructure
    ) -> pd.Series:
        return series_deserialize(value=value, sample=sample)

    @deserialize.register
    def _deserialize_dataframe(
        sample: pd.DataFrame, *, value: JSONStructure
    ) -> pd.DataFrame:
        return dataframe_deserialize(value=value, sample=sample)

    @to_example.register
    def _to_example_numpy(value: np.ndarray) -> str:  # type: ignore
        return numpy_to_example(value=value)

    @to_example.register
    def _to_example_series(value: pd.Series) -> str:
        return series_to_example(value=value)

    @to_example.register
    def _to_example_dataframe(value: pd.DataFrame) -> str:
        return dataframe_to_example(value=value)



"""
Client for talking to arrow endpoints
"""
import logging
import typing as t

# pylint: disable-next=no-name-in-module
from pyarrow._flight import ClientAuthReader, ClientAuthSender  # type: ignore
from pyarrow.flight import ClientAuthHandler, FlightClient  # type: ignore

from taktile_client.arrow.utils import load_certs
from taktile_client.client import Client

from .endpoint import ArrowEndpoints

logger = logging.getLogger(__name__)


class ApiKeyClientAuthHandler(ClientAuthHandler):  # type: ignore
    """Auth Handler"""

    def __init__(self, api_key: str):
        super().__init__()
        self._api_key = api_key

    def authenticate(
        self, outgoing: ClientAuthSender, incoming: ClientAuthReader
    ) -> None:
        """authenticate"""
        outgoing.write(self._api_key)
        self._api_key = incoming.read()

    def get_token(self) -> str:
        """get_token"""
        return self._api_key


class ArrowClient(Client[ArrowEndpoints]):
    """
    Arrow client for talking to models deployed with Taktile.
    """

    def _url_from_host(self, host: str) -> str:
        return f"grpc+tls://{host}:5005"

    def _initalize_endpoints(
        self, url: str, api_key: t.Optional[str]
    ) -> ArrowEndpoints:

        certs = load_certs()
        logger.debug("Performing authentication request against %s", url)
        client = FlightClient(location=url, tls_root_certs=certs)
        if api_key:
            client.authenticate(ApiKeyClientAuthHandler(api_key=api_key))
        actions = client.list_actions()

        return ArrowEndpoints(client=client, actions=actions)



"""
Taktile Client
"""
from .config import arrow_available
from .config import version as __version__

if arrow_available:
    from taktile_client.arrow import ArrowClient

# pylint: disable-next=wrong-import-position, wrong-import-order
from taktile_client.rest import RestClient



"""
Codecs for serializing objects
"""

import typing as t
from datetime import datetime
from decimal import Decimal
from io import BytesIO

import numpy as np
import pyarrow as pa  # type: ignore


class ArbitraryBinaryArray:  # pylint: disable=too-few-public-methods
    """ArbitraryBinaryArray"""

    def __init__(self, val: t.Any) -> None:
        self.val = val

    def __arrow_array__(self) -> pa.Array:
        # convert the underlying array values to a pyarrow Array
        mem_file = BytesIO()
        np.save(mem_file, self.val)  # type: ignore
        return pa.array([bytearray(mem_file.getvalue())], pa.binary())


DESERIALIZE_SEQUENCE_OPS: t.Dict[str, t.Callable[[t.Any], t.Any]] = {
    "SELECT_column": lambda x: getattr(x, "column")(
        DEFAULT_SERIALIZED_COLUMN_NAME
    ),
    "NATIVE_dict": lambda x: x.to_pandas().to_dict("records"),
    "NATIVE_list": lambda x: x.to_pylist(),
    "FINAL_list": list,
    "FINAL_tuple": tuple,
    "FINAL_first": lambda x: x[0],
}

NATIVE_TO_ARROW = {
    int: pa.int8(),
    list: pa.list_,
    float: pa.float64(),
    Decimal: pa.decimal128,
    str: pa.string(),
    dict: pa.struct,
    datetime: pa.timestamp,
}

DEFAULT_SERIALIZED_COLUMN_NAME = "SerializedColumn_gen"



"""
Arrow serialization
"""
import typing as t

import numpy.typing as npt
import pandas as pd  # type: ignore
import pyarrow  # type: ignore

from .factory import deserialize_arrow as deserialize  # noqa: F401
from .factory import serialize_arrow as serialize  # noqa: F401

SerialType = t.Union[pyarrow.Table, pyarrow.Array, pyarrow.Tensor]
UserType = t.Union[npt.NDArray[t.Any], pd.Series, pd.DataFrame]

__all__ = ["deserialize", "serialize"]



"""
Arrow client utilities
"""
import logging
import math
import typing as t

import certifi  # type: ignore
from pyarrow import Table  # type: ignore
from pyarrow.flight import (  # type: ignore
    ActionType,
    FlightClient,
    FlightDescriptor,
    FlightStreamReader,
    FlightStreamWriter,
)

from taktile_client.arrow.serialize import UserType, deserialize, serialize
from taktile_client.config import settings

logger = logging.getLogger(__name__)


def batch_arrow(  # pylint: disable=too-many-locals
    *,
    client: FlightClient,
    action: ActionType,
    payload: UserType,
    nrows: t.Optional[int],
) -> UserType:
    """
    Send a batch via arrow

    Parameters
    ----------
    client : FlightClient
        client to use
    action : ActionType
        action type
    payload : UserType
        payload to send
    nrows: t.Optional[int]
        number of rows to send per batch
    """
    table = serialize(payload)
    batch_size, batch_memory = _get_chunk_size(table, nrows)
    if not (batch_size and batch_memory):
        return None
    descriptor = client.get_flight_info(
        FlightDescriptor.for_command(action.type)
    )
    writer, reader = client.do_exchange(descriptor.descriptor)
    logger.debug(
        "Initiating prediction request with batches of %d records of "
        "~%.2f MB/batch",
        batch_size,
        batch_memory,
    )
    batches = table.to_batches(max_chunksize=batch_size)
    chunks = []
    schema = None
    with writer:
        writer.begin(table.schema)
        for i, batch in enumerate(batches):
            logger.debug("Prediction for batch %d/%d", i + 1, len(batches))
            chunk = _send_batch(
                writer=writer, batch=batch, reader=reader, batch_number=i + 1
            )
            if not chunk:
                continue
            if schema is None and chunk.data.schema is not None:
                schema = chunk.data.schema
            chunks.append(chunk.data)
    deserialized = deserialize(Table.from_batches(chunks, schema))
    return deserialized


def _send_batch(
    writer: FlightStreamWriter,
    batch: t.Any,  # RecordBatch, but there is no python type
    reader: FlightStreamReader,
    batch_number: int,
) -> t.Any:
    try:
        writer.write_batch(batch)
        return reader.read_chunk()
    except Exception as err:  # pylint: disable=broad-except
        logger.error(
            "ERROR: performing prediction for batch %d: %s "
            "The predictions from this batch will be missing from the result",
            batch_number,
            str(err),
        )
        return None


def _get_chunk_size(
    sample_table: Table, batch_size: t.Optional[int]
) -> t.Tuple[t.Optional[int], t.Optional[float]]:

    if sample_table.num_rows == 0:
        logger.error(
            "Empty payload received, "
            "which is currently not supported for arrow endpoints"
        )
        return None, None
    mem_per_record = sample_table.nbytes / sample_table.num_rows
    if batch_size:
        batch_memory_mb = (batch_size * mem_per_record) / 1e6
    else:
        batch_memory_mb = settings.ARROW_BATCH_MB
        batch_size = math.ceil(batch_memory_mb * 1e6 / mem_per_record)
    return batch_size, batch_memory_mb


def load_certs() -> str:
    """Load certificates"""
    with open(certifi.where(), "r", encoding="utf-8") as cert:
        return cert.read()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['taktile_client',
 'taktile_client.arrow',
 'taktile_client.arrow.serialize',
 'taktile_client.rest',
 'taktile_client.rest.serialize']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.8.2,<2.0.0',
 'requests>=2.26.0,<3.0.0',
 'taktile-types>=0.12.0,<1.0.0',
 'tenacity>=8.0.1,<9.0.0']

extras_require = \
{'arrow': ['numpy>=1.21,<2.0',
           'pandas>=1.3,<2.0',
           'pyarrow>=6.0,<7.0',
           'certifi==2021.10.8']}

setup_kwargs = {
    'name': 'taktile-client',
    'version': '1.0.0b5',
    'description': 'A lightweight client to call models deployed on Taktile.',
    'long_description': None,
    'author': 'Taktile GmbH',
    'author_email': 'devops@taktile.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)
